/*-
 *   BSD LICENSE
 *
 *   Copyright(c) 2010-2016 Intel Corporation. All rights reserved.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdbool.h>

#include <rte_common.h>
#include <rte_log.h>
#include <rte_malloc.h>
#include <rte_memory.h>
#include <rte_memcpy.h>
#include <rte_memzone.h>
#include <rte_eal.h>
#include <rte_per_lcore.h>
#include <rte_launch.h>
#include <rte_atomic.h>
#include <rte_cycles.h>
#include <rte_prefetch.h>
#include <rte_lcore.h>
#include <rte_per_lcore.h>
#include <rte_branch_prediction.h>
#include <rte_interrupts.h>
#include <rte_pci.h>
#include <rte_random.h>
#include <rte_debug.h>
#include <rte_ether.h>
#include <rte_ethdev.h>
#include <rte_ring.h>
#include <rte_mempool.h>
#include <rte_mbuf.h>
#include <rte_ip.h>
#include <rte_tcp.h>

#include <assert.h>
#include "rule_trace.h"
#include "hypersplit.h"
#include "bitcuts.h"

static volatile bool force_quit;

#define RTE_LOGTYPE_L2FWD RTE_LOGTYPE_USER1

#define NB_MBUF   8192

#define MAX_PKT_BURST 32
#define BURST_TX_DRAIN_US 100 /* TX drain every ~100us */

/*
 * Configurable number of RX/TX ring descriptors
 */
#define RTE_TEST_RX_DESC_DEFAULT 128
#define RTE_TEST_TX_DESC_DEFAULT 512
static uint16_t nb_rxd = RTE_TEST_RX_DESC_DEFAULT;
static uint16_t nb_txd = RTE_TEST_TX_DESC_DEFAULT;

/* ethernet addresses of ports */
static struct ether_addr l2fwd_ports_eth_addr[RTE_MAX_ETHPORTS];

/* mask of enabled ports */
static uint32_t l2fwd_enabled_port_mask = 0;

/* list of enabled ports */
static uint32_t l2fwd_dst_ports[RTE_MAX_ETHPORTS];

static unsigned int l2fwd_rx_queue_per_lcore = 1;

#define MAX_RX_QUEUE_PER_LCORE 16
#define MAX_TX_QUEUE_PER_PORT 16
struct lcore_queue_conf {
	unsigned n_rx_port;
	unsigned rx_port_list[MAX_RX_QUEUE_PER_LCORE];
} __rte_cache_aligned;
struct lcore_queue_conf lcore_queue_conf[RTE_MAX_LCORE];

static struct rte_eth_dev_tx_buffer *tx_buffer[RTE_MAX_ETHPORTS];

static const struct rte_eth_conf port_conf = {
	.rxmode = {
		.split_hdr_size = 0,
		.header_split   = 0, /**< Header Split disabled */
		.hw_ip_checksum = 0, /**< IP checksum offload disabled */
		.hw_vlan_filter = 0, /**< VLAN filtering disabled */
		.jumbo_frame    = 0, /**< Jumbo Frame Support disabled */
		.hw_strip_crc   = 0, /**< CRC stripped by hardware */
	},
	.txmode = {
		.mq_mode = ETH_MQ_TX_NONE,
	},
};

struct rte_mempool * l2fwd_pktmbuf_pool = NULL;

/* Per-port statistics struct */
struct l2fwd_port_statistics {
	uint64_t tx;
	uint64_t rx;
	uint64_t dropped;
} __rte_cache_aligned;
struct l2fwd_port_statistics port_statistics[RTE_MAX_ETHPORTS];

/* A tsc-based timer responsible for triggering statistics printout */
#define TIMER_MILLISECOND 2000000ULL /* around 1ms at 2 Ghz */
#define MAX_TIMER_PERIOD 86400 /* 1 day max */
static int64_t timer_period = 10 * TIMER_MILLISECOND * 1000; /* default period is 10 seconds */


#define OFF_IPV42PROTO (offsetof(struct ipv4_hdr, next_proto_id))
#define OFF_IPV42PKTID (offsetof(struct ipv4_hdr, packet_id))
#define OFF_IPV42SRCADD (offsetof(struct ipv4_hdr, src_addr))
#define OFF_IPV42DSTADD (offsetof(struct ipv4_hdr, dst_addr))
#define OFF_TCP2SRCPT (offsetof(struct tcp_hdr, src_port))
#define OFF_TCP2DSTPT (offsetof(struct tcp_hdr, dst_port))
#define OFF_ETHHEAD	(sizeof(struct ether_hdr))
#define MBUF_IPV4_2PROTO(m)	\
	rte_pktmbuf_mtod_offset((m), uint8_t *, OFF_ETHHEAD + OFF_IPV42PROTO)
#define PREFETCH_OFFSET 3

enum {
    PC_ALGO_INV = -1,
    PC_ALGO_HYPERSPLIT = 0,
    PC_ALGO_BITCUTS = 1,
    PC_ALGO_MAX = 2
};

enum {
    GRP_ALGO_INV = -1,
    GRP_ALGO_RFG = 0,
    GRP_ALGO_MAX = 1
};


struct platform_config {
    char *s_rule_file;
    int pc_algo;
};

void *result = NULL;

/* Print out statistics on packets dropped */
static void
print_stats(void)
{
	uint64_t total_packets_dropped, total_packets_tx, total_packets_rx;
	unsigned portid;

	total_packets_dropped = 0;
	total_packets_tx = 0;
	total_packets_rx = 0;

	const char clr[] = { 27, '[', '2', 'J', '\0' };
	const char topLeft[] = { 27, '[', '1', ';', '1', 'H','\0' };

		/* Clear screen and move to top left */
	printf("%s%s", clr, topLeft);

	printf("\nPort statistics ====================================");

	for (portid = 0; portid < RTE_MAX_ETHPORTS; portid++) {
		/* skip disabled ports */
		if ((l2fwd_enabled_port_mask & (1 << portid)) == 0)
			continue;
		printf("\nStatistics for port %u ------------------------------"
			   "\nPackets sent: %24"PRIu64
			   "\nPackets received: %20"PRIu64
			   "\nPackets dropped: %21"PRIu64,
			   portid,
			   port_statistics[portid].tx,
			   port_statistics[portid].rx,
			   port_statistics[portid].dropped);

		total_packets_dropped += port_statistics[portid].dropped;
		total_packets_tx += port_statistics[portid].tx;
		total_packets_rx += port_statistics[portid].rx;
	}
	printf("\nAggregate statistics ==============================="
		   "\nTotal packets sent: %18"PRIu64
		   "\nTotal packets received: %14"PRIu64
		   "\nTotal packets dropped: %15"PRIu64,
		   total_packets_tx,
		   total_packets_rx,
		   total_packets_dropped);
	printf("\n====================================================\n");
}

static void
print_packet(struct packet pkt)
{
    int sip1, sip2, sip3, sip4;
    int dip1, dip2, dip3, dip4;
    sip1 = pkt.sip >> 24;
    sip2 = (pkt.sip << 8) >> 24;
    sip3 = (pkt.sip << 16) >> 24;
    sip4 = (pkt.sip << 24) >> 24;
    dip1 = pkt.dip >> 24;
    dip2 = (pkt.dip << 8) >> 24;
    dip3 = (pkt.dip << 16) >> 24;
    dip4 = (pkt.dip << 24) >> 24;

    printf("sip %d.%d.%d.%d, dip %d.%d.%d.%d, ", sip1, sip2, sip3, sip4, dip1, dip2, dip3, dip4);
    printf("sip %"PRIu32" dip %"PRIu32" sport %"PRIu16" dport %"PRIu16", proto %"PRIu8"\n",
            pkt.sip, pkt.dip, pkt.sport, pkt.dport, pkt.proto);
    printf("u64[0]: %"PRIu64", u64[1]: %"PRIu64"\n", pkt.header_u64[0], pkt.header_u64[1]);
    return;
}

static inline void
bc_prepare_one_packet(struct rte_mbuf **pkts_in, struct packet *pkts, int i)
{
    pkts[i].sip = rte_be_to_cpu_32(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint32_t *, OFF_ETHHEAD + OFF_IPV42SRCADD)));
    pkts[i].dip = rte_be_to_cpu_32(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint32_t *, OFF_ETHHEAD + OFF_IPV42DSTADD)));
    /* match_ids[i] = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + OFF_IPV42PKTID))); */
    pkts[i].sport = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + sizeof(struct ipv4_hdr) + OFF_TCP2SRCPT)));
    pkts[i].dport = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + sizeof(struct ipv4_hdr) + OFF_TCP2DSTPT)));
    pkts[i].proto = *(rte_pktmbuf_mtod_offset(pkts_in[i], uint8_t *, OFF_ETHHEAD + OFF_IPV42PROTO));
    pkts[i].header_u64[0] = ((uint64_t)pkts[i].sip << 32) + (uint64_t)pkts[i].dip;
    pkts[i].header_u64[1] = ((uint64_t)pkts[i].sport << 48) + ((uint64_t)pkts[i].dport << 32) + ((uint64_t)pkts[i].proto << 24);
}

static inline void
bc_prepare_packets(struct rte_mbuf **pkts_in, struct packet *pkts, int nb_rx)
{
    int i;

    for (i = 0; i < PREFETCH_OFFSET && i < nb_rx; i++) {
        rte_prefetch0(rte_pktmbuf_mtod(
                    pkts_in[i], void *));
    }

    for (i = 0; i < (nb_rx - PREFETCH_OFFSET); i++) {
        rte_prefetch0(rte_pktmbuf_mtod(
                    pkts_in[i + PREFETCH_OFFSET], void *));
        bc_prepare_one_packet(pkts_in, pkts, i);
    }

    for (; i < nb_rx; i++)
        bc_prepare_one_packet(pkts_in, pkts, i);

}

static inline void
hs_prepare_one_packet(struct rte_mbuf **pkts_in, struct packet *pkts, int i)
{
    pkts[i].dims[0] = rte_be_to_cpu_32(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint32_t *, OFF_ETHHEAD + OFF_IPV42SRCADD)));
    pkts[i].dims[1] = rte_be_to_cpu_32(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint32_t *, OFF_ETHHEAD + OFF_IPV42DSTADD)));
    pkts[i].dims[2] = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + sizeof(struct ipv4_hdr) + OFF_TCP2SRCPT)));
    pkts[i].dims[3] = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + sizeof(struct ipv4_hdr) + OFF_TCP2DSTPT)));
    pkts[i].dims[4] = *(rte_pktmbuf_mtod_offset(pkts_in[i], uint8_t *, OFF_ETHHEAD + OFF_IPV42PROTO));
    pkts[i].match_rule = rte_be_to_cpu_16(*(rte_pktmbuf_mtod_offset(pkts_in[i], uint16_t *, OFF_ETHHEAD + OFF_IPV42PKTID)));
}

static inline void
hs_prepare_packets(struct rte_mbuf **pkts_in, struct packet *pkts, int nb_rx)
{
    int i;

    for (i = 0; i < PREFETCH_OFFSET && i < nb_rx; i++) {
        rte_prefetch0(rte_pktmbuf_mtod(
                    pkts_in[i], void *));
    }

    for (i = 0; i < (nb_rx - PREFETCH_OFFSET); i++) {
        rte_prefetch0(rte_pktmbuf_mtod(
                    pkts_in[i + PREFETCH_OFFSET], void *));
        hs_prepare_one_packet(pkts_in, pkts, i);
    }

    for (; i < nb_rx; i++)
        hs_prepare_one_packet(pkts_in, pkts, i);

}

static inline void
send_one_packet(struct rte_mbuf *m, int res, uint8_t dst_port)
{
    if (likely(res >= 0)) {
	    rte_eth_tx_buffer(dst_port, 0, tx_buffer[dst_port], m);
    }
    else {
        rte_pktmbuf_free(m);
    }
}

static inline void
send_packets(struct rte_mbuf **m, int *match_res, int num, uint8_t tx_portid)
{
	int i;

	/* Prefetch first packets */
	for (i = 0; i < PREFETCH_OFFSET && i < num; i++) {
		rte_prefetch0(rte_pktmbuf_mtod(
				m[i], void *));
	}

	for (i = 0; i < (num - PREFETCH_OFFSET); i++) {
		rte_prefetch0(rte_pktmbuf_mtod(m[
				i + PREFETCH_OFFSET], void *));
        send_one_packet(m[i], match_res[i], tx_portid);
	}

	/* Process left packets */
	for (; i < num; i++)
		send_one_packet(m[i], match_res[i], tx_portid);
}

/* BitCuts main processing loop */
static void
bc_l2fwd_main_loop(void)
{
	struct rte_mbuf *pkts_burst[MAX_PKT_BURST];
	struct rte_mbuf *m;
	int sent;
	unsigned lcore_id;
	unsigned i, j, portid, nb_rx;
	struct lcore_queue_conf *qconf;
	struct rte_eth_dev_tx_buffer *buffer;

    int id;
    unsigned dst_port;

	lcore_id = rte_lcore_id();
	qconf = &lcore_queue_conf[lcore_id];

	if (qconf->n_rx_port == 0) {
		RTE_LOG(INFO, L2FWD, "lcore %u has nothing to do\n", lcore_id);
		return;
	}

	RTE_LOG(INFO, L2FWD, "entering main loop on lcore %u\n", lcore_id);

	for (i = 0; i < qconf->n_rx_port; i++) {

		portid = qconf->rx_port_list[i];
		RTE_LOG(INFO, L2FWD, " -- lcoreid=%u portid=%u\n", lcore_id,
			portid);

	}

    struct packet pkts[MAX_PKT_BURST];
    /* int match_ids[MAX_PKT_BURST]; */
    int match_res[MAX_PKT_BURST];

    while (!force_quit) {

		/*
		 * Read packet from RX queues
		 */
		for (i = 0; i < qconf->n_rx_port; i++) {

			portid = qconf->rx_port_list[i];
			nb_rx = rte_eth_rx_burst((uint8_t) portid, 0,
						 pkts_burst, MAX_PKT_BURST);
            dst_port = l2fwd_dst_ports[portid];

            if (nb_rx > 0) {

                bc_prepare_packets(pkts_burst, pkts, nb_rx);

                for (j = 0; j < nb_rx; j++) {
                    match_res[j] = bc_search_tmp(&result, &pkts[j]);
                }

                send_packets(pkts_burst, match_res, nb_rx, dst_port);
            }
		}
	}
}

/* HyperSplit main processing loop */
static void
hs_l2fwd_main_loop(void)
{
	struct rte_mbuf *pkts_burst[MAX_PKT_BURST];
	struct rte_mbuf *m;
	int sent;
	unsigned lcore_id;
	int i, j, nb_rx;
    unsigned portid;
	struct lcore_queue_conf *qconf;
	struct rte_eth_dev_tx_buffer *buffer;

    int id;
    unsigned dst_port;

	lcore_id = rte_lcore_id();
	qconf = &lcore_queue_conf[lcore_id];

	if (qconf->n_rx_port == 0) {
		RTE_LOG(INFO, L2FWD, "lcore %u has nothing to do\n", lcore_id);
		return;
	}

	RTE_LOG(INFO, L2FWD, "entering main loop on lcore %u\n", lcore_id);

	for (i = 0; i < qconf->n_rx_port; i++) {

		portid = qconf->rx_port_list[i];
		RTE_LOG(INFO, L2FWD, " -- lcoreid=%u portid=%u\n", lcore_id,
			portid);

	}

    struct packet pkts[MAX_PKT_BURST];
    /* int match_ids[MAX_PKT_BURST]; */
    int match_res[MAX_PKT_BURST];

    while (!force_quit) {

		/*
		 * Read packet from RX queues
		 */
		for (i = 0; i < qconf->n_rx_port; i++) {

			portid = qconf->rx_port_list[i];
			nb_rx = rte_eth_rx_burst((uint8_t) portid, 0,
						 pkts_burst, MAX_PKT_BURST);
            dst_port = l2fwd_dst_ports[portid];

            if (nb_rx > 0) {

                hs_prepare_packets(pkts_burst, pkts, nb_rx);

                for (j = 0; j < nb_rx; j++) {
                    match_res[j] = hs_search_tmp(&result, &pkts[j]);
                }

                send_packets(pkts_burst, match_res, nb_rx, dst_port);
            }
		}
	}
}

static int
l2fwd_launch_one_lcore(__attribute__((unused)) void *dummy)
{
    struct platform_config *p_plat_cfg = (struct platform_config *)dummy;

    switch (p_plat_cfg->pc_algo) {
        case PC_ALGO_BITCUTS:
            bc_l2fwd_main_loop();
            break;

        case PC_ALGO_HYPERSPLIT:
            hs_l2fwd_main_loop();
            break;

        default:
            return -ENOTSUP;
    }
    
	return 0;
}

/* display usage */
static void
l2fwd_usage(const char *prgname)
{
	printf("%s [EAL options] -- -p PORTMASK [-q NQ]\n"
	       "  -p PORTMASK: hexadecimal bitmask of ports to configure\n"
	       "  -q NQ: number of queue (=ports) per lcore (default is 1)\n"
		   "  -T PERIOD: statistics will be refreshed each PERIOD seconds (0 to disable, 10 default, 86400 maximum)\n",
	       prgname);
}

static int
l2fwd_parse_portmask(const char *portmask)
{
	char *end = NULL;
	unsigned long pm;

	/* parse hexadecimal string */
	pm = strtoul(portmask, &end, 16);
	if ((portmask[0] == '\0') || (end == NULL) || (*end != '\0'))
		return -1;

	if (pm == 0)
		return -1;

	return pm;
}

static unsigned int
l2fwd_parse_nqueue(const char *q_arg)
{
	char *end = NULL;
	unsigned long n;

	/* parse hexadecimal string */
	n = strtoul(q_arg, &end, 10);
	if ((q_arg[0] == '\0') || (end == NULL) || (*end != '\0'))
		return 0;
	if (n == 0)
		return 0;
	if (n >= MAX_RX_QUEUE_PER_LCORE)
		return 0;

	return n;
}

static int
l2fwd_parse_timer_period(const char *q_arg)
{
	char *end = NULL;
	int n;

	/* parse number string */
	n = strtol(q_arg, &end, 10);
	if ((q_arg[0] == '\0') || (end == NULL) || (*end != '\0'))
		return -1;
	if (n >= MAX_TIMER_PERIOD)
		return -1;

	return n;
}

/* Parse the argument given in the command line of the application */
static int
l2fwd_parse_args(struct platform_config *p_plat_cfg, int argc, char **argv)
{
	int opt, ret;
	char **argvopt;
	int option_index;
	char *prgname = argv[0];
	static struct option lgopts[] = {
		{NULL, 0, 0, 0}
	};

	argvopt = argv;

	while ((opt = getopt_long(argc, argvopt, "p:q:T:r:a:",
				  lgopts, &option_index)) != EOF) {

		switch (opt) {
		/* portmask */
		case 'p':
			l2fwd_enabled_port_mask = l2fwd_parse_portmask(optarg);
			if (l2fwd_enabled_port_mask == 0) {
				printf("invalid portmask\n");
				l2fwd_usage(prgname);
				return -1;
			}
			break;

		/* nqueue */
		case 'q':
			l2fwd_rx_queue_per_lcore = l2fwd_parse_nqueue(optarg);
			if (l2fwd_rx_queue_per_lcore == 0) {
				printf("invalid queue number\n");
				l2fwd_usage(prgname);
				return -1;
			}
			break;

		/* timer period */
		case 'T':
			timer_period = l2fwd_parse_timer_period(optarg) * 1000 * TIMER_MILLISECOND;
			if (timer_period < 0) {
				printf("invalid timer period\n");
				l2fwd_usage(prgname);
				return -1;
			}
			break;

        case 'r':
            p_plat_cfg->s_rule_file = optarg;
            break;

        case 'a':
            if (!strcmp(optarg, "hs")) {
                p_plat_cfg->pc_algo = PC_ALGO_HYPERSPLIT;
            
            } else if (!strcmp(optarg, "bc")) {
                p_plat_cfg->pc_algo = PC_ALGO_BITCUTS;
            }

            break;

		/* long options */
		case 0:
			l2fwd_usage(prgname);
			return -1;

		default:
			l2fwd_usage(prgname);
			return -1;
		}
	}

	if (optind >= 0)
		argv[optind-1] = prgname;

	ret = optind-1;
	optind = 0; /* reset getopt lib */
	return ret;
}

/* Check the link status of all ports in up to 9s, and print them finally */
static void
check_all_ports_link_status(uint8_t port_num, uint32_t port_mask)
{
#define CHECK_INTERVAL 100 /* 100ms */
#define MAX_CHECK_TIME 90 /* 9s (90 * 100ms) in total */
	uint8_t portid, count, all_ports_up, print_flag = 0;
	struct rte_eth_link link;

	printf("\nChecking link status");
	fflush(stdout);
	for (count = 0; count <= MAX_CHECK_TIME; count++) {
		if (force_quit)
			return;
		all_ports_up = 1;
		for (portid = 0; portid < port_num; portid++) {
			if (force_quit)
				return;
			if ((port_mask & (1 << portid)) == 0)
				continue;
			memset(&link, 0, sizeof(link));
			rte_eth_link_get_nowait(portid, &link);
			/* print link status if flag set */
			if (print_flag == 1) {
				if (link.link_status)
					printf("Port %d Link Up - speed %u "
						"Mbps - %s\n", (uint8_t)portid,
						(unsigned)link.link_speed,
				(link.link_duplex == ETH_LINK_FULL_DUPLEX) ?
					("full-duplex") : ("half-duplex\n"));
				else
					printf("Port %d Link Down\n",
						(uint8_t)portid);
				continue;
			}
			/* clear all_ports_up flag if any link down */
			if (link.link_status == ETH_LINK_DOWN) {
				all_ports_up = 0;
				break;
			}
		}
		/* after finally printing all link status, get out */
		if (print_flag == 1)
			break;

		if (all_ports_up == 0) {
			printf(".");
			fflush(stdout);
			rte_delay_ms(CHECK_INTERVAL);
		}

		/* set the print_flag if all ports up or timeout */
		if (all_ports_up == 1 || count == (MAX_CHECK_TIME - 1)) {
			print_flag = 1;
			printf("done\n");
		}
	}
}

static void
signal_handler(int signum)
{
	if (signum == SIGINT || signum == SIGTERM) {
		printf("\n\nSignal %d received, preparing to exit...\n",
				signum);
		force_quit = true;
	}
}

static uint64_t f_make_timediff(const struct timespec stop,
        const struct timespec start)
{
    return (stop.tv_sec * 1000000ULL + stop.tv_nsec / 1000)
        - (start.tv_sec * 1000000ULL + start.tv_nsec / 1000);
}

static int f_build(int pc_algo, void *built_result,
        const struct partition *p_pa)
{
    assert(pc_algo > PC_ALGO_INV && pc_algo < PC_ALGO_MAX);
    assert(built_result && p_pa && p_pa->subsets && p_pa->rule_num > 1);
    assert(p_pa->subset_num > 0 && p_pa->subset_num <= PART_MAX);

    switch (pc_algo) {
    case PC_ALGO_HYPERSPLIT:
        return hs_build(built_result, p_pa);
        break;

    case PC_ALGO_BITCUTS:
        return bc_build(built_result, p_pa);
        break;

    default:
        *(typeof(built_result) *)built_result = NULL;
        return -ENOTSUP;
    }
}


int
main(int argc, char **argv)
{
	struct lcore_queue_conf *qconf;
	struct rte_eth_dev_info dev_info;
	int ret;
	uint8_t nb_ports;
	uint8_t nb_ports_available;
	uint8_t portid, last_port;
	unsigned lcore_id, rx_lcore_id;
	unsigned nb_ports_in_mask = 0;
        
    struct timespec starttime, stoptime;
    uint64_t timediff;
    struct partition pa;
    struct platform_config plat_cfg = {
        .s_rule_file = NULL,
        .pc_algo = PC_ALGO_INV,
    };

	/* init EAL */
	ret = rte_eal_init(argc, argv);
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "Invalid EAL arguments\n");
	argc -= ret;
	argv += ret;

	force_quit = false;
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);

	/* parse application arguments (after the EAL ones) */
	ret = l2fwd_parse_args(&plat_cfg, argc, argv);
	if (ret < 0)
		rte_exit(EXIT_FAILURE, "Invalid L2FWD arguments\n");

    /* Loading classifier */
    if (plat_cfg.s_rule_file) {
        pa.subsets = calloc(1, sizeof(*pa.subsets));
        if (!pa.subsets) {
            perror("Cannot allocate memory for subsets");
            exit(-1);
        }

        if (load_rules(pa.subsets, plat_cfg.s_rule_file)) {
            exit(-1);
        }

        pa.subset_num = 1;
        pa.rule_num = pa.subsets[0].rule_num;

        printf("Building\n");

        clock_gettime(CLOCK_MONOTONIC, &starttime);

        if (f_build(plat_cfg.pc_algo, &result, &pa)) {
            printf("Building fail\n");
            exit(-1);
        }

        clock_gettime(CLOCK_MONOTONIC, &stoptime);

        printf("Building pass\n");
        printf("Time for building: %"PRIu64"(us)\n",
                f_make_timediff(stoptime, starttime));

    }

	/* create the mbuf pool */
	l2fwd_pktmbuf_pool = rte_pktmbuf_pool_create("mbuf_pool", NB_MBUF, 32,
		0, RTE_MBUF_DEFAULT_BUF_SIZE, rte_socket_id());
	if (l2fwd_pktmbuf_pool == NULL)
		rte_exit(EXIT_FAILURE, "Cannot init mbuf pool\n");

	nb_ports = rte_eth_dev_count();
	if (nb_ports == 0)
		rte_exit(EXIT_FAILURE, "No Ethernet ports - bye\n");

	if (nb_ports > RTE_MAX_ETHPORTS)
		nb_ports = RTE_MAX_ETHPORTS;

	/* reset l2fwd_dst_ports */
	for (portid = 0; portid < RTE_MAX_ETHPORTS; portid++)
		l2fwd_dst_ports[portid] = 0;
	last_port = 0;

	/*
	 * Each logical core is assigned a dedicated TX queue on each port.
	 */
	for (portid = 0; portid < nb_ports; portid++) {
		/* skip ports that are not enabled */
		if ((l2fwd_enabled_port_mask & (1 << portid)) == 0)
			continue;

		if (nb_ports_in_mask % 2) {
			l2fwd_dst_ports[portid] = last_port;
			l2fwd_dst_ports[last_port] = portid;
		}
		else
			last_port = portid;

		nb_ports_in_mask++;

		rte_eth_dev_info_get(portid, &dev_info);
	}
	if (nb_ports_in_mask % 2) {
		printf("Notice: odd number of ports in portmask.\n");
		l2fwd_dst_ports[last_port] = last_port;
	}

	rx_lcore_id = 0;
	qconf = NULL;

	/* Initialize the port/queue configuration of each logical core */
	for (portid = 0; portid < nb_ports; portid++) {
		/* skip ports that are not enabled */
		if ((l2fwd_enabled_port_mask & (1 << portid)) == 0)
			continue;

		/* get the lcore_id for this port */
		while (rte_lcore_is_enabled(rx_lcore_id) == 0 ||
		       lcore_queue_conf[rx_lcore_id].n_rx_port ==
		       l2fwd_rx_queue_per_lcore) {
			rx_lcore_id++;
			if (rx_lcore_id >= RTE_MAX_LCORE)
				rte_exit(EXIT_FAILURE, "Not enough cores\n");
		}

		if (qconf != &lcore_queue_conf[rx_lcore_id])
			/* Assigned a new logical core in the loop above. */
			qconf = &lcore_queue_conf[rx_lcore_id];

		qconf->rx_port_list[qconf->n_rx_port] = portid;
		qconf->n_rx_port++;
		printf("Lcore %u: RX port %u\n", rx_lcore_id, (unsigned) portid);
	}

	nb_ports_available = nb_ports;

	/* Initialise each port */
	for (portid = 0; portid < nb_ports; portid++) {
		/* skip ports that are not enabled */
		if ((l2fwd_enabled_port_mask & (1 << portid)) == 0) {
			printf("Skipping disabled port %u\n", (unsigned) portid);
			nb_ports_available--;
			continue;
		}
		/* init port */
		printf("Initializing port %u... ", (unsigned) portid);
		fflush(stdout);
		ret = rte_eth_dev_configure(portid, 1, 1, &port_conf);
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "Cannot configure device: err=%d, port=%u\n",
				  ret, (unsigned) portid);

		rte_eth_macaddr_get(portid,&l2fwd_ports_eth_addr[portid]);

		/* init one RX queue */
		fflush(stdout);
		ret = rte_eth_rx_queue_setup(portid, 0, nb_rxd,
					     rte_eth_dev_socket_id(portid),
					     NULL,
					     l2fwd_pktmbuf_pool);
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "rte_eth_rx_queue_setup:err=%d, port=%u\n",
				  ret, (unsigned) portid);

		/* init one TX queue on each port */
		fflush(stdout);
		ret = rte_eth_tx_queue_setup(portid, 0, nb_txd,
				rte_eth_dev_socket_id(portid),
				NULL);
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "rte_eth_tx_queue_setup:err=%d, port=%u\n",
				ret, (unsigned) portid);

		/* Initialize TX buffers */
		tx_buffer[portid] = rte_zmalloc_socket("tx_buffer",
				RTE_ETH_TX_BUFFER_SIZE(MAX_PKT_BURST), 0,
				rte_eth_dev_socket_id(portid));
		if (tx_buffer[portid] == NULL)
			rte_exit(EXIT_FAILURE, "Cannot allocate buffer for tx on port %u\n",
					(unsigned) portid);

		rte_eth_tx_buffer_init(tx_buffer[portid], MAX_PKT_BURST);

		ret = rte_eth_tx_buffer_set_err_callback(tx_buffer[portid],
				rte_eth_tx_buffer_count_callback,
				&port_statistics[portid].dropped);
		if (ret < 0)
				rte_exit(EXIT_FAILURE, "Cannot set error callback for "
						"tx buffer on port %u\n", (unsigned) portid);

		/* Start device */
		ret = rte_eth_dev_start(portid);
		if (ret < 0)
			rte_exit(EXIT_FAILURE, "rte_eth_dev_start:err=%d, port=%u\n",
				  ret, (unsigned) portid);

		printf("done: \n");

		rte_eth_promiscuous_enable(portid);

		printf("Port %u, MAC address: %02X:%02X:%02X:%02X:%02X:%02X\n\n",
				(unsigned) portid,
				l2fwd_ports_eth_addr[portid].addr_bytes[0],
				l2fwd_ports_eth_addr[portid].addr_bytes[1],
				l2fwd_ports_eth_addr[portid].addr_bytes[2],
				l2fwd_ports_eth_addr[portid].addr_bytes[3],
				l2fwd_ports_eth_addr[portid].addr_bytes[4],
				l2fwd_ports_eth_addr[portid].addr_bytes[5]);

		/* initialize port stats */
		memset(&port_statistics, 0, sizeof(port_statistics));
	}

	if (!nb_ports_available) {
		rte_exit(EXIT_FAILURE,
			"All available ports are disabled. Please set portmask.\n");
	}

	check_all_ports_link_status(nb_ports, l2fwd_enabled_port_mask);

	ret = 0;
	/* launch per-lcore init on every lcore */
	/* rte_eal_mp_remote_launch(l2fwd_launch_one_lcore, NULL, CALL_MASTER); */
    rte_eal_mp_remote_launch(l2fwd_launch_one_lcore, (void *)(&plat_cfg), CALL_MASTER);
	RTE_LCORE_FOREACH_SLAVE(lcore_id) {
		if (rte_eal_wait_lcore(lcore_id) < 0) {
			ret = -1;
			break;
		}
	}

	for (portid = 0; portid < nb_ports; portid++) {
		if ((l2fwd_enabled_port_mask & (1 << portid)) == 0)
			continue;
		printf("Closing port %d...", portid);
		rte_eth_dev_stop(portid);
		rte_eth_dev_close(portid);
		printf(" Done\n");
	}
	printf("Bye...\n");

	return ret;
}
